import gettext
from ikabot.config import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.naval import getAvailableShips

t = gettext.translation('buyResources', localedir, languages=languages, fallback=True)
_ = t.gettext

def getCityMilitaryData(session, city_id):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city_id : int
    
    returns
    -------
    data : dict
    """
    params = {
        "view": "cityMilitary",
        "activeTab": "tabUnits",
        "cityId": city_id,
        "currentTab": "tabUnits",
        "currentCityId": city_id,
        "templateView": "cityMilitary",
        "actionRequest": actionRequest,
        "ajax": "1"
    }
    data = session.post(params=params)
    data = json.loads(data, strict=False)
    return data[1][1][1]

def extractTooltipsAndValues(data):
    tooltips = re.findall(r'<div class="tooltip">(.*?)</div>', data)
    values = re.findall(r'<td>\s*([\d.-]+)\s*</td>', data)
    return tooltips, values

def calculateTotals(tooltips, values):
    total_units = 0
    total_ships = 0

    desc_value_dict = {}

    for i, (tooltip, value) in enumerate(zip(tooltips, values)):
        desc_value_dict.setdefault(tooltip, []).append(value)

        if value.isdigit() and i <= 14:
            total_units += int(value)
        elif value.isdigit():
            total_ships += int(value)

    return desc_value_dict, total_units, total_ships

def getArmyAvailable(session, type_army, destination_city_id, origin_city_id, event):
    params = {
        "view": "deployment",
        "deploymentType": "army" if type_army else "fleet",
        "destinationCityId": destination_city_id,
        "backgroundView": "city",
        "currentCityId": origin_city_id,
        "actionRequest": actionRequest,
        "ajax": 1
    }

    data = session.post(params=params)
    amount_results = re.findall(r'<div class=\\"amount\\">(.*?)<\\/div>', data)
    
    if type_army:
        army_results = re.findall(r'name=\\"cargo_army_([^\\]+)_upkeep\\"\\n\s+value=\\"([^\\"]+)\\"', data)
        weight_total_ships = int(getAvailableShips(session)) * 500 if type_army else 0
        weight_results = re.findall(r'<div class=\\"weight\\">(.*?)<\\/div>', data)
    else:
        army_results = re.findall(r'name=\\"cargo_fleet_([^\\]+)_upkeep\\"\\n\s+value=\\"([^\\"]+)\\"', data)


    army_available = {}
    weight_total_army = 0

    if army_results:
        for i, result in enumerate(army_results):
            army_code = result[0]
            army = "cargo_{}_{}_upkeep".format("army" if type_army else "fleet", army_code)
            army_only = "cargo_{}_{}".format("army" if type_army else "fleet", army_code)
            upkeep = result[1]
            quantity = amount_results[i]
            army_available[army] = upkeep
            army_available[army_only] = quantity
            if type_army and weight_results and int(weight_results[i]) > 0:
                weight_total_army += int(quantity) * int(weight_results[i])
    
        if type_army and weight_total_army > weight_total_ships:
            banner()
            print('Not enough ships to transport all the units!')
            enter()
            return None
        return army_available
    return None


def sendArmy(session, origin_city, destination_city, type_army, army_available):
    params = {
        "action": "transportOperations",
        "function": "deployArmy" if type_army else "deployFleet",
        "actionRequest": actionRequest,
        "islandId": destination_city['islandId'],
        "destinationCityId": destination_city['id'],
        "deploymentType": "army" if type_army else "fleet",
        "backgroundView": "city",
        "currentCityId": origin_city['id'],
        "templateView": "deployment",
        "ajax": 1
    }

    for army in army_available:
        params[army] = army_available[army]

    session.post(params=params)

def stationArmy(session,event, stdin_fd, predetermined_input):
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    type_army = True
    try:
        banner()

        ids, cities = getIdsOfCities(session)
        army = {}
        print('Total:')
        print('{:>19}|{:>19}|{:>19}|'.format('', 'Units', 'Ships'))

        for city_id in cities:
            city = cities[city_id]
            data = getCityMilitaryData(session, city['id'])
            desc, values = extractTooltipsAndValues(data)
            army, total_units, total_ships = calculateTotals(desc, values)
            
            print('{:>19}|{:>19}|{:>19}|'.format(city['name'], total_units, total_ships))
        
        print()
        print(_('(0) Back'))
        print(_('(1) Move troops'))
        print(_('(2) Move ships'))
        print(_('(3) Move all ground units to a city.'))
        print(_('(4) Move all maritime units to a city.'))
        print(_('(5) Move all units to a city.'))
        
        selected = read(min=0, max=5, digit=True)
        if selected == 0:
            event.set()
            return
        elif selected in (1, 2):
            print('Origin city:')
            origin_city = chooseCity(session)
            print()
            print('Destination city:')
            destination_city = chooseCity(session)
            if origin_city['id'] == destination_city['id']:
                banner()
                print('The city of origin and the destination city cannot be the same!')
                enter()
                event.set()
            else:
                type_army = selected == 1
                army_available = getArmyAvailable(session, type_army, destination_city['id'], origin_city['id'], event)
                if army_available != None:
                    sendArmy(session, origin_city, destination_city, type_army, army_available)
                    print('Army sent!')
                    enter()
                    event.set()
                else:
                    print()
                    print('No {} units available in {}.'.format('ground' if type_army else 'maritime', origin_city['name']))
                    enter()
                    event.set()
        elif selected in (3,4,5):
            print('Destination city:')
            destination_city = chooseCity(session)
            ids, cities = getIdsOfCities(session)
            
            if selected in (3,5):
                type_army = True
                for city_id in cities:
                    if city_id != destination_city['id']:
                        city = cities[city_id]
                        army_available = getArmyAvailable(session, type_army, destination_city['id'], city['id'], event)
                        if army_available != None:
                            sendArmy(session, city, destination_city, type_army, army_available)
                        else:
                            print('No ground units available in {}.'.format(city['name']))
            if selected in (4,5):
                type_army = False
                for city_id in cities:
                    if city_id != destination_city['id']:
                        city = cities[city_id]
                        army_available = getArmyAvailable(session, type_army, destination_city['id'], city['id'], event)
                        if army_available != None:
                            sendArmy(session, city, destination_city, type_army, army_available)
                        else:
                            print('No maritime units available in {}.'.format(city['name']))
            enter()
            event.set()               
    except KeyboardInterrupt:
        event.set()
        return  


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import gettext
import multiprocessing
import time
import datetime
from ikabot.config import *
from ikabot.web.session import *
from ikabot.helpers.gui import *
from ikabot.function.donate import donate
from ikabot.function.update import update
from ikabot.helpers.pedirInfo import read
from ikabot.function.getStatus import getStatus
from ikabot.function.donationBot import donationBot
from ikabot.helpers.botComm import updateTelegramData, telegramDataIsValid
from ikabot.helpers.process import updateProcessList
from ikabot.function.constructionList import constructionList
from ikabot.function.searchForIslandSpaces import searchForIslandSpaces
from ikabot.function.alertAttacks import alertAttacks
from ikabot.function.vacationMode import vacationMode
from ikabot.function.activateMiracle import activateMiracle
from ikabot.function.trainArmy import trainArmy
from ikabot.function.sellResources import sellResources
from ikabot.function.checkForUpdate import checkForUpdate
from ikabot.function.distributeResources import distributeResources
from ikabot.function.alertLowWine import alertLowWine
from ikabot.function.buyResources import buyResources
from ikabot.function.loginDaily import loginDaily
from ikabot.function.sendResources import sendResources
from ikabot.function.constructBuilding import constructBuilding
from ikabot.function.shipMovements import shipMovements
from ikabot.function.importExportCookie import importExportCookie
from ikabot.function.autoPirate import autoPirate
from ikabot.function.investigate import investigate
from ikabot.function.attackBarbarians import attackBarbarians
from ikabot.function.proxyConf import proxyConf, show_proxy
from ikabot.function.killTasks import killTasks
from ikabot.function.decaptchaConf import decaptchaConf
from ikabot.function.dumpWorld import dumpWorld
from ikabot.function.stationArmy import stationArmy
from ikabot.function.logs import logs


t = gettext.translation('command_line', localedir, languages=languages, fallback=True)
_ = t.gettext


def menu(session, checkUpdate=True):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    checkUpdate : bool
    """
    if checkUpdate:
        checkForUpdate()

    show_proxy(session)

    banner()

    process_list = updateProcessList(session)
    if len(process_list) > 0:
        # Insert table header
        table = process_list.copy()
        table.insert(0,{'pid':'pid', 'action':'task','date':'date','status':'status'})
        # Get max length of strings in each category (date is always going to be 15)
        maxPid, maxAction, maxStatus = [max(i) for i in [[len(str(r['pid'])) for r in table], [len(str(r['action'])) for r in table], [len(str(r['status'])) for r in table]]]
        # Print header
        print('|{:^{maxPid}}|{:^{maxAction}}|{:^15}|{:^{maxStatus}}|'.format(table[0]['pid'], table[0]['action'], table[0]['date'], table[0]['status'], maxPid=maxPid, maxAction=maxAction, maxStatus=maxStatus))
        # Print process list
        [print('|{:^{maxPid}}|{:^{maxAction}}|{:^15}|{:^{maxStatus}}|'.format(r['pid'], r['action'], datetime.datetime.fromtimestamp(r['date']).strftime('%b %d %H:%M:%S'), r['status'], maxPid=maxPid, maxAction=maxAction, maxStatus=maxStatus)) for r in process_list]
        print('')

    menu_actions = {
        1:            constructionList,
        2:            sendResources,
        3:            distributeResources,
        4:            getStatus,
        5:            searchForIslandSpaces,
        6:            loginDaily,
        101:            alertAttacks,
        102:            alertLowWine,
        111:            buyResources,
        112:            sellResources,
        121:            donate,
        122:            donationBot,
        10:            vacationMode,
        11:            activateMiracle,
        131:            trainArmy,
        132:            stationArmy,
        13:            shipMovements,
        14:            constructBuilding,
        15:            update,
        16:            importExportCookie,
        17:            autoPirate,
        18:            investigate,
        19:            attackBarbarians,
        20:            dumpWorld,
        141:            proxyConf,
        142:            updateTelegramData,
        143:            killTasks,
        144:            decaptchaConf,
        145:            logs,
                    }

    print(_('(0)  Exit'))
    print(_('(1)  Construction list'))
    print(_('(2)  Send resources'))
    print(_('(3)  Distribute resources'))
    print(_('(4)  Account status'))
    print(_('(5)  Monitor islands'))
    print(_('(6)  Login daily'))
    print(_('(7)  Alerts / Notifications'))
    print(_('(8)  Marketplace'))
    print(_('(9)  Donate'))
    print(_('(10) Activate vacation mode'))
    print(_('(11) Activate miracle'))
    print(_('(12) Military actions'))
    print(_('(13) See movements'))
    print(_('(14) Construct building'))
    print(_('(15) Update Ikabot'))
    print(_('(16) Import / Export cookie'))
    print(_('(17) Auto-Pirate'))
    print(_('(18) Investigate'))
    print(_('(19) Attack barbarians'))
    print(_('(20) Dump / View world'))
    print(_('(21) Options / Settings'))
    total_options = len(menu_actions) + 1
    selected = read(min=0, max=total_options, digit=True)
    
    if selected == 7:
        banner()
        print(_('(0) Back'))
        print(_('(1) Alert attacks'))
        print(_('(2) Alert wine running out'))

        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 100
            
    if selected == 8:
        banner()
        print(_('(0) Back'))
        print(_('(1) Buy resources'))
        print(_('(2) Sell resources'))

        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 110
            
    if selected == 9:
        banner()
        print(_('(0) Back'))
        print(_('(1) Donate once'))
        print(_('(2) Donate automatically'))

        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 120
            
    if selected == 12:
        banner()
        print(_('(0) Back'))
        print(_('(1) Train Army'))
        print(_('(2) Send Troops/Ships'))
        selected = read(min=0, max=2, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 130

    if selected == 21:
        banner()
        print(_('(0) Back'))
        print(_('(1) Configure Proxy'))
        if telegramDataIsValid(session):
            print(_('(2) Change the Telegram data'))
        else:
            print(_('(2) Enter the Telegram data'))
        print(_('(3) Kill tasks'))
        print(_('(4) Configure captcha resolver'))
        print(_('(5) Logs'))

        selected = read(min=0, max=5, digit=True)
        if selected == 0:
            menu(session)
            return
        if selected > 0:
            selected += 140
        


    if selected != 0:
        try:
            event = multiprocessing.Event()  # creates a new event
            process = multiprocessing.Process(target=menu_actions[selected], args=(session, event, sys.stdin.fileno(), config.predetermined_input), name=menu_actions[selected].__name__)
            process.start()
            process_list.append({'pid': process.pid, 'action': menu_actions[selected].__name__, 'date': time.time(), 'status': 'started'})
            updateProcessList(session, programprocesslist=process_list)
            event.wait()  # waits for the process to fire the event that's been given to it. When it does  this process gets back control of the command line and asks user for more input
        except KeyboardInterrupt:
            pass
        menu(session, checkUpdate=False)
    else:
        if isWindows:
            # in unix, you can exit ikabot and close the terminal and the processes will continue to execute
            # in windows, you can exit ikabot but if you close the terminal, the processes will die
            print(_('Closing this console will kill the processes.'))
            enter()
        clear()
        os._exit(0)  # kills the process which executes this statement, but it does not kill it's child processes


def init():
    home = 'USERPROFILE' if isWindows else 'HOME'
    os.chdir(os.getenv(home))
    if not os.path.isfile(ikaFile):
        open(ikaFile, 'w')
        os.chmod(ikaFile, 0o600)


def start():
    init()
    config.has_params = len(sys.argv) > 1
    for arg in sys.argv:
        try:
            config.predetermined_input.append(int(arg))
        except ValueError:
            config.predetermined_input.append(arg)
    config.predetermined_input.pop(0)

    session = Session()
    try:
        menu(session)
    finally:
        clear()
        session.logout()


def main():
    manager = multiprocessing.Manager()
    predetermined_input = manager.list()
    config.predetermined_input = predetermined_input
    try:
        start()
    except KeyboardInterrupt:
        clear()


if __name__ == '__main__':
    # On Windows calling this function is necessary.
    if sys.platform.startswith('win'):
        multiprocessing.freeze_support()
    main()


#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import gettext
import traceback
import sys
import threading
import json
import gzip
import re
import os
import ast
import ikabot.config as config
from pathlib import Path
from ikabot.config import *
from ikabot.helpers.gui import enter, banner, bcolors
from ikabot.helpers.varios import wait, getDateTime
from ikabot.helpers.botComm import sendToBot
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.pedirInfo import read
from ikabot.helpers.getJson import getIsland
from ikabot.helpers.process import set_child_mode

t = gettext.translation('dumpWorld', localedir, languages=languages, fallback=True)
_ = t.gettext

LINE_UP = '\033[1A'
LINE_CLEAR = '\x1b[2K'
#              status, history, start_time
stop_updating = threading.Event()
lock = threading.Lock()
shared_data = ['','',0,stop_updating, lock]
home = 'USERPROFILE' if isWindows else 'HOME'
selected_islands = set()


def dumpWorld(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input

    try:
        banner()
        if os.path.exists(os.getenv(home) + '/ikabot_world_dumps'):
            print('1) Create new dump')
            print('2) Load existing dump')
            choice = read(min=1, max=2, digit=True)
            if choice == 2:
                view_dump(session, event)
                event.set()
                return
        banner()
        print('{}⚠️ BEWARE - THE RESULTING DUMP CONTAINS ACCOUNT IDENTIFYING INFORMATION ⚠️{}\n'.format(bcolors.WARNING, bcolors.ENDC))
        print('This action will take a couple of hours to complete. Are you sure you want to initiate a data dump now? (Y|N)')
        choice = read(values=['y', 'Y', 'n', 'N'])
        if (choice in ['n','N']):
            event.set()
            return
        print('Type in the waiting time between each request in miliseconds (default = 1500): ')
        choice = read(min=0, max=10000, digit=True, default=1500)
        waiting_time = int(choice)/1000
        print('Start scan form island id (0 to start from beginning) (default = 0): ')
        choice = read(min=0, digit=True, default=0)
        start_id = int(choice)
        print('Do you want only shallow data about the islands? If yes you will not be able to search the dump by player names but the dump will be quick. (Y|N): ')
        choice = read(values=['y', 'Y', 'n', 'N'])
        shallow = choice
        
        thread = threading.Thread(target=update_terminal, args=(shared_data,))
        thread.start()
        set_child_mode(session)
        info = _('\nDumped world data\n')
        setInfoSignal(session, info)

        dump_path = do_it(session, waiting_time, start_id, shallow)

        shared_data[3].set()
        shared_data[4].acquire()
        time.sleep(5)

        banner()
        print('\n{}SUCCESS!{} World data has been dumped to {} in {}s \n'.format(bcolors.GREEN, bcolors.ENDC,dump_path, str(round(time.time()-shared_data[2]))))
        enter()
        event.set()
        return
    except Exception:
        shared_data[3].set()
        shared_data[4].acquire(timeout=10)
        event.set()
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
        return


def do_it(session, waiting_time, start_id, shallow):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
        Session object
    waiting_time : int
        Time to wait between each network request (to prevent getting rate limited)
    start_id : int
        Id of the island to start the dump from (0 starts from beginning)
    shallow : str
        String that determines if the data should be shallow (only map accessible data)
    
    Returns
    -------
    dump_path: str
    """

    shared_data[2] = time.time()
    world = {'name': 's' + str(session.mundo) + '-' + str(session.servidor),
             'self_name': session.username,
             'dump_start_date': time.time(),
             'dump_end_date': 0,
             'islands': [],
             'shallow': shallow in ['y', 'Y']
            }
    shared_data.append(world)
    #scan 0 to 50 x and y
    shallow_islands = []
    update_status('Initiating first map sweep', 0, 0, True)
    update_status('Getting (0-50,0-50) islands', 25, 1.25)
    data = session.post('action=WorldMap&function=getJSONArea&x_min=0&x_max=50&y_min=0&y_max=50')
    for x, val in json.loads(data)['data'].items():
        for y, val2 in val.items():
            shallow_islands.append({'x': x, 'y': y, 'id': val2[0], 'name': val2[1], 'resource_type': val2[2], 'miracle_type': val2[3], 'wood_lvl': val2[6], 'players': val2[7] })
    update_status('Getting (50-100,0-50) islands', 50, 2.5)
    time.sleep(0.5)
    data = session.post('action=WorldMap&function=getJSONArea&x_min=50&x_max=100&y_min=0&y_max=50')
    for x, val in json.loads(data)['data'].items():
        for y, val2 in val.items():
            shallow_islands.append({'x': x, 'y': y, 'id': val2[0], 'name': val2[1], 'resource_type': val2[2], 'miracle_type': val2[3], 'wood_lvl': val2[6], 'players': val2[7] })
    update_status('Getting (0-50,50-100) islands', 75, 3.75)
    time.sleep(0.5)
    data = session.post('action=WorldMap&function=getJSONArea&x_min=0&x_max=50&y_min=50&y_max=100')
    for x, val in json.loads(data)['data'].items():
        for y, val2 in val.items():
            shallow_islands.append({'x': x, 'y': y, 'id': val2[0], 'name': val2[1], 'resource_type': val2[2], 'miracle_type': val2[3], 'wood_lvl': val2[6], 'players': val2[7] })
    update_status('Getting (50-100,50-100) islands', 100, 5)
    time.sleep(0.5)
    data = session.post('action=WorldMap&function=getJSONArea&x_min=50&x_max=100&y_min=50&y_max=100')
    for x, val in json.loads(data)['data'].items():
        for y, val2 in val.items():
            shallow_islands.append({'x': x, 'y': y, 'id': val2[0], 'name': val2[1], 'resource_type': val2[2], 'miracle_type': val2[3], 'wood_lvl': val2[6], 'players': val2[7] })

# [
# "58",         //id 0
# "Phytios",    //name 1
# "1",          //resource type 2
# "2",          //type of miracle 3
# "5",          // ?? 4
# "4",          // ?? 5
# "9",          // lumber level  6
# "12",         // number of people 7
# 0,            // piracy in range 8
# "0",          // helios tower 9
# "0",          // red 10
# "0"           // blue 11
# ]
    
    dump_path = os.getenv(home) + '/ikabot_world_dumps/s' + str(session.mundo) + '-' + str(session.servidor) + '/'
    dump_path = dump_path.replace('\\','/')
    dump_name = getDateTime() + '.json.gz'

    if shallow in ['y','Y']:
        dump_name = dump_name.replace('.json.gz', '_shallow') + '.json.gz'
        update_status('Shallow dump is on. Dumping data...', 100, 100, True)
        world['islands'] = shallow_islands
        dump(world, dump_path, dump_name)
        return dump_path + dump_name

    all_island_ids = set()
    total_settlements = 0
    for island in shallow_islands:
        all_island_ids.add(island['id'])
        total_settlements += int(island['players'])

    update_status('Got {} islands with {} towns in total'.format(len(all_island_ids), str(total_settlements)), 100, 5, True)
    update_status('Getting data for each island. This will take a while...', 0, 5, True)

    #scan each island

    world_islands_number = len(all_island_ids)
    all_island_ids = list(split(sorted(map(int, all_island_ids)), 1))
    
    dump_islands(shared_data, all_island_ids[0], waiting_time, start_id, session, world_islands_number)

    update_status('Got {} individual islands'.format(world_islands_number), 100, 100, True)

    update_status('Dumping data to {}'.format(dump_path + dump_name), 100, 100, True)

    dump(shared_data[5], dump_path, dump_name)
    

    return dump_path + dump_name

def split(a, n):
    k, m = divmod(len(a), n)
    return (a[i*k+min(i, m):(i+1)*k+min(i+1, m)] for i in range(n))

def dump(world, dump_path, dump_name):
    world['dump_end_date'] = time.time()
    p = Path(dump_path)
    p.mkdir(exist_ok=True, parents=True)
    with gzip.open(dump_path+dump_name,'wb') as file:
        json_string = json.dumps(world).encode('utf-8')
        file.write(json_string)
       
def dump_islands(shared_data, all_island_ids, waiting_time, start_id, session, world_islands_number):
    for i, island_id in enumerate(sorted(map(int, all_island_ids))):
        if(int(island_id) < start_id):
            continue
        update_status('Getting island id {}'.format(island_id), len(shared_data[5]['islands'])/world_islands_number*100, 5+(len(shared_data[5]['islands'])/world_islands_number * 95))
        html = ''
        try:
            html = session.get('view=island&islandId=' + str(island_id))
        except Exception:
            # try again
            html = session.get('view=island&islandId=' + str(island_id))
        island = getIsland(html)
        shared_data[4].acquire()
        shared_data[5]['islands'].append(island)
        shared_data[4].release()
        time.sleep(waiting_time)

def update_terminal(shared_data):
    while(True):
        banner()
        print('\n')
        shared_data[4].acquire()
        print(shared_data[1])
        shared_data[4].release()
        chars = ['\\','|','/','─']
        for i in range(20):
            shared_data[4].acquire()
            print(' '*120, end='\r')
            print(shared_data[0] + ',\tdt: ' + str(round(time.time()-shared_data[2],2)) + 's\t' + chars[i%4] , end='\r')
            shared_data[4].release()
            time.sleep(0.05)
        if stop_updating.is_set():
            return
        
def update_status(message, percent, percent_total, add_history = False):
    shared_data[4].acquire()
    shared_data[0] = message + '\t' + str(round(percent,1)) + '%,\ttotal: ' + str(round(percent_total,1)) + '%'
    if(add_history):
        shared_data[1] += shared_data[0] + '\n'
    shared_data[4].release()
    

def view_dump(session, event):
    files = [file.replace('\\','/') for file in get_files(os.getenv(home) + '/ikabot_world_dumps') if '.json.gz' in file ]

    print('All dumps are stored in ' + os.getenv(home) + '/ikabot_world_dumps\n')
    print('Choose a dump to view:')
    for i, file in enumerate(files):
        print(str(i) + ') ' + file.split('/')[-2] + ' ' + file.split('/')[-1].replace('.json.gz','').replace('_',' '))
    choice = read(min = 0, max = len(files)-1, digit=True)
    print('Loading dump...')
    selected_dump = files[choice]
    with gzip.open(selected_dump, 'rb') as file:
        selected_dump = json.load(file)
    
    while True:
        banner()
        print_map(selected_dump['islands'])
        print('0) Back')
        print('1) Search islands by island criteria')
        if not selected_dump['shallow']:
            print('2) Search islands by player name')
            print('3) Search for nearest inactive players')

        choice = read(min=0, max=3, digit=True)
        if choice == 0:
            event.set()
            return
        elif choice == 1:
            print('Search island by a certain criteria. The available properties of an island are:')
            print('resource_type : [1,2,3,4] // these are  Wine, Marble, Cristal, Sulfur')
            print('miracle_type : [1,2,3,4,5,6,7,8] // hephaistos forge is number 5')
            print('wood_lvl : [1..] // this is the forest level on the island')
            print('players : [0..] // number of players on the island')
            print('ex. If I wanted to find all islands with less than 10 players and forest level 30 with hephaistos I would type in:')
            print('players < 10 and wood_lvl == 30 and miracle_type == 5\n')
            condition = read(msg="Enter the condition: ")

            try:
                filtered_islands = [island for island in filter(lambda x: filter_on_condition(x, condition), selected_dump['islands'] if selected_dump['shallow'] else convert_to_shallow(selected_dump['islands']))]
            except (SyntaxError, KeyError):
                print('Condition is bad, please use only the available island properties and use python standard conditional sytnax (and, or, <, >, ==, (, ), etc... )')
                enter()
                continue
            
            print('The satisfying islands are:')
            [print(island) for island in filtered_islands]
            enter()
        elif choice == 2:
            if selected_dump['shallow']:
                print('You can not search by player name because this dump is shallow and doesn\'t contain data about players!')
                enter()
                continue
            player_name = read(msg='Type in the player name: ')
            # search for players by name
            players = []
            for island in selected_dump['islands']:
                for city in island['cities']:
                    if city['type'] != 'empty' and player_name in city['Name']:
                        players.append((player_name, island['id']))
            # return if none are found
            if not len(players):
                print('No players found!')
                enter()
                continue
            # select one 
            print('Chose a player to add to selection: ')
            for i, player in enumerate(unique_tuples(players)):
                print(str(i) + ') ' + player[0])
            choice = read(min=0, max=len(list(unique_tuples(players)))-1, digit=True)
            # add his islands to selection
            for player in players:
                if player[0] == list(unique_tuples(players))[choice][0]:
                    selected_islands.add(int(player[1]))
        elif choice == 3:
            if selected_dump['shallow']:
                print('You can not search by player name because this dump is shallow and doesn\'t contain data about players!')
                enter()
                continue
            coords = read(msg='Type in a center point (x,y): ').replace('(','').replace(')','').split(',')
            coords = (int(coords[0]), int(coords[1]))
            number_of_inactives = read(msg='How many inactives should be displayed? (min=1, default=25): ', min = 1, digit=True, default=25)
            #sort islands based on distance from center point
            islands_sorted = sorted(selected_dump['islands'], key=lambda island: ((island['x'] - coords[0]) ** 2 + (island['y'] - coords[1]) ** 2) ** 0.5)
            print('The nearest 25 inactive players are: ')
            #below follows the unholiest way to get the first n cities which are contained in an island object which is contained in a list of islands without duplicates in one line of code using python list comprehension
            seen = set()
            inactives = [city for island in islands_sorted for city in island['cities'] if city['type'] != 'empty' and city['state'] == 'inactive' and isinstance([(seen.add(city['Name']),) if city['Name'] not in seen else None][0],tuple)][:number_of_inactives]
            for i, city in enumerate(inactives):
                print(str(i+1) + ') ' + city['Name'])
            enter()


def print_map(islands):
    """Prints out a 100x100 matrix with all world islands on it. Selected islands are colored red.
    Parameters
    ----------
    islands : [object]
        List of island objects to be displayed
    """

    map = [[bcolors.DARK_BLUE + '██' + bcolors.ENDC for j in range(100)] for i in range(100)] # 100x100 matrix of dark blue ██
    selected_island_coords = []

    for island in islands:
        if int(island['id']) in selected_islands:
            map[int(island['x'])-1][int(island['y'])-1] = bcolors.DARK_RED + '◉ ' + bcolors.ENDC 
            selected_island_coords.append((int(island['x']),int(island['y'])))
        else:
            map[int(island['x'])-1][int(island['y'])-1] = bcolors.DARK_GREEN + '◉ ' + bcolors.ENDC

    for row in reversed(map):
        print(''.join(row))

    print(bcolors.DARK_BLUE + '██' + bcolors.ENDC + ' - Water, ' + \
          bcolors.DARK_GREEN + '◉' + bcolors.ENDC + ' - Island, '+ \
          bcolors.DARK_RED + '◉' + bcolors.ENDC + ' - Selected\n'
        )

    print('Selected islands: ' + str(selected_island_coords))
    

def filter_on_condition(island, condition):
    """Returns true if island satisfies condition
    Parameters
    ----------
    island : object
        Island to be tested on condition
    condition : str
        String that represents a valid python condition to be applied to filter the list of islands

    Returns
    -------
    is_satisfied : bool
        Bool indicating whether or not the island object satisfies the condition
    """

    condition = ast.parse(condition)
    for node in ast.walk(condition):
        if isinstance(node, ast.Compare):
            left = node.left.id
            right = node.comparators[0].n if isinstance(node.comparators[0], ast.Num) else node.comparators[0].id
            op = node.ops[0]
            if op.__class__ == ast.Lt:
                if not int(island[left]) < int(right):
                    return False
            elif op.__class__ == ast.Gt:
                if not int(island[left]) > int(right):
                    return False
            elif op.__class__ == ast.Eq:
                if not int(island[left]) == int(right):
                    return False
    return True

    

def convert_to_shallow(islands):
    """Converts a list of islands from a deep dump into a shallow dump version of that list
    Parameters
    ----------
    islands : [object]
        List of island objects to be converted
    
    Returns
    -------
    islands : [object]
        List of objects that represent the stripped-down version of an island 
    """
    return [{'x': str(island['x']), 'y': str(island['y']), 'id': island['id'], 'name': island['name'], 'resource_type': island['tradegood'], 'miracle_type': island['wonder'], 'wood_lvl': island['resourceLevel'], 'players': len([city for city in island['cities'] if city['type'] != 'empty'])} for island in islands]

def unique_tuples(tuples):
    """Iterates over tuples with a unique first element
    """
    seen = {}
    for t in tuples:
        if t[0] not in seen:
            seen[t[0]] = True
            yield t

def get_files(path):
    """
    Returns all full paths to every file in a directory and all it's subdirectories
    Parameters
    ----------
    path : str
        Path to directory
    Returns
    -------
    files : list
    """
    files = []
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            files.append(os.path.join(dirpath, filename))
    return files



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import math
import json
import gettext
import traceback
from decimal import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.varios import addThousandSeparator
from ikabot.helpers.gui import enter, banner
from ikabot.helpers.getJson import getCity
from ikabot.helpers.signals import setInfoSignal
from ikabot.helpers.planRoutes import waitForArrival
from ikabot.helpers.pedirInfo import getIdsOfCities, read
from ikabot.helpers.naval import getTotalShips
from ikabot.config import *
from ikabot.helpers.botComm import *
from ikabot.helpers.resources import *
from ikabot.helpers.market import *


t = gettext.translation('buyResources', localedir, languages=languages, fallback=True)
_ = t.gettext


def chooseResource(session, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    """
    print(_('Which resource do you want to buy?'))
    for index, material_name in enumerate(materials_names):
        print('({:d}) {}'.format(index+1, material_name))
    choise = read(min=1, max=5)
    resource = choise - 1
    if resource == 0:
        resource = 'resource'
    data = {
        'cityId': city['id'],
        'position': city['pos'],
        'view': 'branchOffice',
        'activeTab': 'bargain',
        'type': 444,
        'searchResource': resource,
        'range': city['rango'],
        'backgroundView': 'city',
        'currentCityId': city['id'],
        'templateView': 'branchOffice',
        'currentTab': 'bargain',
        'actionRequest': actionRequest,
        'ajax': 1
    }
    # this will set the chosen resource in the store
    session.post(params=data)
    resource = choise - 1
    # return the chosen resource
    return resource


def getOffers(session, city):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    Returns
    -------
    offers : list[dict]
    """
    html = getMarketHtml(session, city)
    hits = re.findall(r'short_text80">(.*?) *<br/>\((.*?)\)\s *</td>\s *<td>(\d+)</td>\s *<td>(.*?)/td>\s *<td><img src="(.*?)\.png[\s\S]*?white-space:nowrap;">(\d+)\s[\s\S]*?href="\?view=takeOffer&destinationCityId=(\d+)&oldView=branchOffice&activeTab=bargain&cityId=(\d+)&position=(\d+)&type=(\d+)&resource=(\w+)"', html)
    offers = []
    for hit in hits:
        offer = {
            'ciudadDestino': hit[0],
            'jugadorAComprar': hit[1],
            'bienesXminuto': int(hit[2]),
            'amountAvailable': int(hit[3].replace(',', '').replace('.', '').replace('<', '')),
            'tipo': hit[4],
            'precio': int(hit[5]),
            'destinationCityId': hit[6],
            'cityId': hit[7],
            'position': hit[8],
            'type': hit[9],
            'resource': hit[10]
        }
        
        #Parse CDN Images to material type
        if offer["tipo"] == '//gf2.geo.gfsrv.net/cdn19/c3527b2f694fb882563c04df6d8972':
             offer["tipo"] = 'wood'
        elif offer["tipo"] == '//gf1.geo.gfsrv.net/cdnc6/94ddfda045a8f5ced3397d791fd064':
            offer["tipo"] = 'wine'     
        elif  offer["tipo"] == '//gf3.geo.gfsrv.net/cdnbf/fc258b990c1a2a36c5aeb9872fc08a':
             offer["tipo"] = 'marble'
        elif  offer["tipo"] == '//gf2.geo.gfsrv.net/cdn1e/417b4059940b2ae2680c070a197d8c':
             offer["tipo"] = 'glass'
        elif  offer["tipo"] == '//gf1.geo.gfsrv.net/cdn9b/5578a7dfa3e98124439cca4a387a61':
             offer["tipo"] = 'sulfur'
        else:
            continue
            
        offers.append(offer)
    return offers


def calculateCost(offers, total_amount_to_buy):
    """
    Parameters
    ----------
    offers : list[dict]
    total_amount_to_buy : int
    Returns
    -------
    total_cost : int
    """
    total_cost = 0
    for offer in offers:
        if total_amount_to_buy == 0:
            break
        buy_amount = min(offer['amountAvailable'], total_amount_to_buy)
        total_amount_to_buy -= buy_amount
        total_cost += buy_amount * offer['precio']
    return total_cost


def chooseCommertialCity(commercial_cities):
    """
    Parameters
    ----------
    commercial_cities : list[dict]

    Returns
    -------
    commercial_city : dict
    """
    print(_('From which city do you want to buy resources?\n'))
    for i, city in enumerate(commercial_cities):
        print('({:d}) {}'.format(i + 1, city['name']))
    selected_city_index = read(min=1, max=len(commercial_cities))
    return commercial_cities[selected_city_index - 1]


def buyResources(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()

        # get all the cities with a store
        commercial_cities = getCommercialCities(session)
        if len(commercial_cities) == 0:
            print(_('There is no store build'))
            enter()
            event.set()
            return

        # choose which city to buy from
        if len(commercial_cities) == 1:
            city = commercial_cities[0]
        else:
            city = chooseCommertialCity(commercial_cities)
            banner()

        # choose resource to buy
        resource = chooseResource(session, city)
        banner()

        # get all the offers of the chosen resource from the chosen city
        offers = getOffers(session, city)
        if len(offers) == 0:
            print(_('There are no offers available.'))
            enter()
            event.set()
            return

        # display offers to the user
        total_price = 0
        total_amount = 0
        for offer in offers:
            amount = offer['amountAvailable']
            price = offer['precio']
            cost = amount * price
            print(_('amount:{}').format(addThousandSeparator(amount)))
            print(_('price :{:d}').format(price))
            print(_('cost  :{}').format(addThousandSeparator(cost)))
            print('')
            total_price += cost
            total_amount += amount

        # ask how much to buy
        print(_('Total amount available to purchase: {}, for {}').format(addThousandSeparator(total_amount), addThousandSeparator(total_price)))
        available = city['freeSpaceForResources'][resource]
        if available < total_amount:
            print(_('You just can buy {} due to storing capacity').format(addThousandSeparator(available)))
            total_amount = available
        print('')
        amount_to_buy = read(msg=_('How much do you want to buy?: '), min=0, max=total_amount)
        if amount_to_buy == 0:
            event.set()
            return

        # calculate the total cost
        (gold, __) = getGold(session, city)
        total_cost = calculateCost(offers, amount_to_buy)

        print(_('\nCurrent gold: {}.\nTotal cost  : {}.\nFinal gold  : {}.'). format(addThousandSeparator(gold), addThousandSeparator(total_cost), addThousandSeparator(gold - total_cost)))
        print(_('Proceed? [Y/n]'))
        rta = read(values=['y', 'Y', 'n', 'N', ''])
        if rta.lower() == 'n':
            event.set()
            return

        print(_('It will be purchased {}').format(addThousandSeparator(amount_to_buy)))
        enter()
    except KeyboardInterrupt:
        event.set()
        return

    set_child_mode(session)
    event.set()

    info = _('\nI will buy {} from {} to {}\n').format(addThousandSeparator(amount_to_buy), materials_names[resource], city['cityName'])
    setInfoSignal(session, info)
    try:
        do_it(session, city, offers, amount_to_buy)
    except Exception as e:
        msg = _('Error in:\n{}\nCause:\n{}').format(info, traceback.format_exc())
        sendToBot(session, msg)
    finally:
        session.logout()


def buy(session, city, offer, amount_to_buy, ships_available):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    offer : dict
    amount_to_buy : int
    """
    ships = int(math.ceil((Decimal(amount_to_buy) / Decimal(500))))
    data_dict = {
        'action': 'transportOperations',
        'function': 'buyGoodsAtAnotherBranchOffice',
        'cityId': offer['cityId'],
        'destinationCityId': offer['destinationCityId'],
        'oldView': 'branchOffice',
        'position': city['pos'],
        'avatar2Name': offer['jugadorAComprar'],
        'city2Name': offer['ciudadDestino'],
        'type': int(offer['type']),
        'activeTab': 'bargain',
        'transportDisplayPrice': 0,
        'premiumTransporter': 0,
        'normalTransportersMax': ships_available,
        'capacity': 5,
        'max_capacity': 5,
        'jetPropulsion': 0,
        'transporters': ships,
        'backgroundView': 'city',
        'currentCityId': offer['cityId'],
        'templateView': 'takeOffer',
        'currentTab': 'bargain',
        'actionRequest': actionRequest,
        'ajax': 1
    }
    url = 'view=takeOffer&destinationCityId={}&oldView=branchOffice&activeTab=bargain&cityId={}&position={}&type={}&resource={}&backgroundView=city&currentCityId={}&templateView=branchOffice&actionRequest={}&ajax=1'.format(offer['destinationCityId'], offer['cityId'], offer['position'], offer['type'], offer['resource'], offer['cityId'], actionRequest)
    data = session.post(url)
    html = json.loads(data, strict=False)[1][1][1]
    hits = re.findall(r'"tradegood(\d)Price"\s*value="(\d+)', html)
    for hit in hits:
        data_dict['tradegood{}Price'.format(hit[0])] = int(hit[1])
        data_dict['cargo_tradegood{}'.format(hit[0])] = 0
    hit = re.search(r'"resourcePrice"\s*value="(\d+)', html)
    if hit:
        data_dict['resourcePrice'] = int(hit.group(1))
        data_dict['cargo_resource'] = 0
    resource = offer['resource']
    if resource == 'resource':
        data_dict['cargo_resource'] = amount_to_buy
    else:
        data_dict['cargo_tradegood{}'.format(resource)] = amount_to_buy
    session.post(params=data_dict)
    msg = _('I buy {} to {} from {}').format(addThousandSeparator(amount_to_buy), offer['ciudadDestino'], offer['jugadorAComprar'])
    sendToBotDebug(session, msg, debugON_buyResources)


def do_it(session, city, offers, amount_to_buy):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    city : dict
    offers : list[dict]
    amount_to_buy : int
    """
    while True:
        for offer in offers:
            if amount_to_buy == 0:
                return
            if offer['amountAvailable'] == 0:
                continue

            ships_available = waitForArrival(session)
            storageCapacity = ships_available * 500
            buy_amount = min(amount_to_buy, storageCapacity, offer['amountAvailable'])

            amount_to_buy -= buy_amount
            offer['amountAvailable'] -= buy_amount
            buy(session, city, offer, buy_amount, ships_available)
            # start from the beginning again, so that we always buy from the cheapest offers fisrt
            break



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import gettext
import traceback
from ikabot.config import *
from ikabot.helpers.gui import *
from ikabot.helpers.varios import *
from ikabot.helpers.botComm import *
from ikabot.helpers.pedirInfo import *
from ikabot.helpers.process import set_child_mode
from ikabot.helpers.getJson import getCity
from ikabot.helpers.signals import setInfoSignal

t = gettext.translation('investigate', localedir, languages=languages, fallback=True)
_ = t.gettext

def get_studies(session):
    html = session.get()
    city = getCity(html)
    city_id = city['id']
    url = 'view=researchAdvisor&oldView=updateGlobalData&cityId={0}&backgroundView=city&currentCityId={0}&templateView=researchAdvisor&actionRequest={1}&ajax=1'.format(city_id, actionRequest)
    resp = session.post(url)
    resp = json.loads(resp, strict=False)
    return resp[2][1]


def study(session, studies, num_study):
    html = session.get()
    city = getCity(html)
    city_id = city['id']
    research_type = studies['js_researchAdvisorChangeResearchType{}'.format(num_study)]['ajaxrequest'].split('=')[-1]
    url = 'action=Advisor&function=doResearch&actionRequest={}&type={}&backgroundView=city&currentCityId={}&templateView=researchAdvisor&ajax=1'.format(actionRequest, research_type, city_id)
    session.post(url)

def experiment(session, experiments):
     while experiments['qty'] > 0:
        # Validate if material is still there..oterwhise log it and send it via bot
        session.get('view=city&cityId={}'.format(experiments["cityID"]), noIndex=True)
        data = session.get("view=updateGlobalData&ajax=1", noIndex=True)
        json_data = json.loads(data, strict=False)
        json_data = json_data[0][1]['headerData']
        current_glass = int(json_data['currentResources']['3'])

        if (current_glass < 300000):
            sendToBot(session, f"Experiment process ended on {experiments['cityName']} due lack of glass ({addThousandSeparator(current_glass)})")
            break

        url = f'action=CityScreen&function=buyResearch&cityId={experiments["cityID"]}&position={experiments["pos"]}&backgroundView=city&currentCityId={experiments["cityID"]}&templateView=academy&actionRequest={actionRequest}&ajax=1'
        session.post(url)
        experiments['qty'] = experiments['qty'] - 1
        sendToBot(session, f"Experiment done on {experiments['cityName']}, left = {experiments['qty']} time (s)")

        # Terminate it if no of experiments = 0
        if (experiments['qty'] == 0):
            break

        # Every 4h, added 1m extra
        time.sleep(241 * 60)


def investigate(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    try:
        banner()
        
        print('\nSelect an option:')
        print('1) Study')
        print('2) Conduct experiment')
        option = read(min=1, max=2)

        if option == 1:
            studies = get_studies(session)
            keys = list(studies.keys())
            num_studies = len([key for key in keys if 'js_researchAdvisorChangeResearchTypeTxt' in key])

            available = []
            for num_study in range(num_studies):
                if 'js_researchAdvisorProgressTxt{}'.format(num_study) in studies:
                    available.append(num_study)

            if len(available) == 0:
                print(_('There are no available studies.'))
                enter()
                event.set()
                return

            print(_('Which one do you wish to study?'))
            print('0) None')
            for index, num_study in enumerate(available):
                print('{:d}) {}'.format(index+1, studies['js_researchAdvisorNextResearchName{}'.format(num_study)]))
            choice = read(min=0, max=len(available))

            if choice == 0:
                event.set()
                return

            study(session, studies, available[choice-1])
            print(_('Done.'))
            enter()
            event.set()
        else:
            # Experiment
            experiments = {}
            total_glass = 0
            found_academy = -1

            #while (total_glass < 300000 or found_academy < 0):
            banner()
            print('Pick city: ')
            city = chooseCity(session)
            total_glass = int(city['recursos'][3])
            
            # Check if enough glass
            if (total_glass < 300000 ):
                print(f'Not enough glass ({addThousandSeparator(total_glass)}), try another city. Min=300k')
                time.sleep(2)
                enter()
                event.set()
                return
            
            # Search for Academy
            for building in city['position']:
                if building['building'] == 'academy':
                    found_academy = building['position']

            if (found_academy < 0):
                print(f'No academy in this town, pick another one')
                time.sleep(2)
                enter()
                event.set()
                return
            
            max_experiments = (total_glass // 300000)
            banner()
            print(f'How many experiments? Min=1, Max={max_experiments}')
            choice = read(min=1, max=max_experiments)

            # Build experiments dict
            experiments['cityID'] = city['id']
            experiments['cityName'] = city['name']
            experiments['pos'] = found_academy
            experiments['qty'] = choice

            # Process
            set_child_mode(session)
            event.set()

            info = (f'Process: Experiments\n\nWill excecute {choice} times every 4h')

            try:
                sendToBot(session, info)
                experiment(session, experiments)
            except Exception as e:
                error_msg = f'Error in:\n{info}\nCause:\n{traceback.format_exc()}'
                sendToBot(session, error_msg)
            finally:
                session.logout()

    except KeyboardInterrupt:
        event.set()
        return



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
import gettext
import sys
import pprint
import requests
import base64
import gzip
import io
from ikabot.helpers.pedirInfo import read, enter
from ikabot.helpers.gui import *
from ikabot.config import *

t = gettext.translation('logs', localedir, languages=languages, fallback=True)
_ = t.gettext

key = 'Cy8JNToyOz0zUltBJFo3CiIgKyc1NwMyP1kJEzUjJFgzJQsGOAAJCz5EOFY='

def logs(session, event, stdin_fd, predetermined_input):
    """
    Parameters
    ----------
    session : ikabot.web.session.Session
    event : multiprocessing.Event
    stdin_fd: int
    predetermined_input : multiprocessing.managers.SyncManager.list
    """
    sys.stdin = os.fdopen(stdin_fd)
    config.predetermined_input = predetermined_input
    global key
    key = base64.b64decode(key).decode("utf-8")
    key = "".join([chr(ord(key[i]) ^ ord('ikabot'[i % len('ikabot')])) for i in range(len(key))])
    try:
        while True:
            banner()
            print("0) Back")
            print("1) Set log level")
            print("2) View logs")
            choice = read(min = 0, max = 2, digit = True)
            if choice == 0:
                event.set()
                return
            elif choice == 1:
                banner()
                print("The current log level is: " + logLevelsText[session.logLevel] + "\n")
                print("0) DEBUG")
                print("1) INFO")
                print("2) WARN")
                print("3) ERROR")
                choice = read(min = 0, max = 3, digit = True)
                session.updateLogLevel(choice)
                print("The log level has been set to: " + logLevelsText[session.logLevel] + "\n")
                enter()
            else:    
                viewLogs(session)
                    
      
    except KeyboardInterrupt:
        event.set()
        return
    

def viewLogs(session, sort = '-date', page = 0):
    while True:

        banner()
        logs = session.getLogs(sort = sort, page = page)

        displayLogs = [
                        '[' + bcolors.RED + 'ERROR' + bcolors.ENDC + ']' + ' ' + log['date'] + ' ' + log['message'][:50].replace('\n','') if log['level'] == 3 else \
                        '[' + bcolors.WARNING + 'WARN ' + bcolors.ENDC + ']' + ' ' + log['date'] + ' ' + log['message'][:50].replace('\n','') if log['level'] == 2 else \
                        '[' + bcolors.BLUE + 'INFO ' + bcolors.ENDC + ']' + ' ' + log['date'] + ' ' + log['message'][:50].replace('\n','') if log['level'] == 1 else \
                        '[' + bcolors.GREEN + 'DEBUG' + bcolors.ENDC + ']' + ' ' + log['date'] + ' ' + log['message'][:50].replace('\n','') \
                        for log in logs ]

        printChoiceList(displayLogs)
        print("Type in the number of the log you wish to see the details of. Press enter for next page, or type in back for previous page.")
        print("To sort this list type in the property by which to sort: 'level', 'date', 'message'")
        choice = read(min = 1, max = 25, additionalValues =  ["back", "level", "date", "message", ""], empty = True)
        if str(choice).isdigit():
            viewLogDetails(session, page, choice, sort)
            viewLogs(session, sort=sort, page=page)
        elif choice == '':
            viewLogs(session, sort=sort, page=page+1)
        elif choice in ['level', 'date', 'message']:
            choice = '-'+choice if choice == sort else choice   # add - to sort so it reverses order if user typed it twice
            viewLogs(session, sort=choice, page=0)
        elif choice == 'back' and (sort != 'date' or page != 0):
            return
        else:
            pass
        

def viewLogDetails(session, page, log_number, sort):
    try:
        log = session.getLogs(page=page, sort=sort)[log_number-1]
        pprint.pprint(log, indent=4)
        print('\nPress [Enter] to upload this log dump to pastebin. Or type in back to go back\
               \nPlease beware that this log might contain identifiable information such as city, player and island ids and names.')
        choice = read(empty = True, values='back')
        if choice == 'back':
            return
        elif choice == '':
            pass
        choice = read(msg="Do you wish to ommit the request history from the log before uploading it to pastebin? [Y|n]: ", values=['n','N','Y','y'])
        if choice in ['y','Y']:
            log['request_history'] = 'None'
        print('Posting to pastebin...')
        print(post_to_pastebin(session, log))
        enter()
    except KeyboardInterrupt:
        pass

def post_to_pastebin(session, log):
    try:                                                                                                                                             

        log['request_history'] = compress_str(log['request_history'])
        response = requests.post("https://pastebin.com/api/api_post.php", data={"api_dev_key": base64.b64decode(key).decode("utf-8"), "api_option": "paste", "api_paste_code": pprint.pformat(log, indent=4)})
        assert response.status_code == 200, "Access code is not 200: {}\n text is: {}".format(response.status_code, response.text)
        return 'Access this log at ' + response.text
    except:
        return 'Error, paste failed!'
    
def compress_str(str):
    """Compresses string str and returns base64 value
    """
    s_bytes = str.encode('utf-8')
    out = io.BytesIO()
    with gzip.GzipFile(fileobj=out, mode='w') as f:
        f.write(s_bytes)
    return base64.b64encode(out.getvalue()).decode('utf-8')

def decompress_str(str):
    """Takes in a compressed string in base64 form and decompresses and decodes it
        Use this function to decode the request history that is pasted in the pastebin!
    """
    s_bytes = base64.b64decode(str)
    inp = io.BytesIO(s_bytes)
    with gzip.GzipFile(fileobj=inp, mode='r') as f:
        decompressed_bytes = f.read()
    return decompressed_bytes.decode('utf-8')